//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) - 2013 - Scilab Enterprises - Vladislav Trubkin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
//

//Description of a Model FMU

function description = fmiModelDescription(model)
    [lhs,rhs]=argn()
    if rhs <> 1 then
        error(999, msprintf(_("%s: Wrong number of input arguments: %d expected.\n"), "fmiModelDescription", 1));
    elseif typeof(model) <> "FMU" then
        error(999, msprintf(_("%s: Wrong type for input argument #%d: %s expected.\n"), "fmiModelDescription", 1, "FMU"));
    end
        if ~isempty(model.descriptionFMU) then
            description = model.descriptionFMU;
            return;
        end
    //read xml
    xmlDescription = xmlRead(model.xmlFile)
    if typeof(xmlDescription) <> "XMLDoc" then
       error(msprintf(_("%s: The file %s doesn''t exist or is not read accessible.\n"),"fmiModelDescription", "modelDescription.xml"));
    end
    //Size of UnitDefinitions
    positionOfUnitDef = find(xmlDescription.root.children.name == "UnitDefinitions")
    if positionOfUnitDef <> [] then
       fmiUnitDefSize = xmlDescription.root.children(positionOfUnitDef).children.size
    else
       fmiUnitDefSize = 0;
    end
    //Size of TypeDefinitions
    positionOfTypeDef = find(xmlDescription.root.children.name == "TypeDefinitions")
    if positionOfTypeDef <> [] then
       fmiTypeDefSize = xmlDescription.root.children(positionOfTypeDef).children.size
    else
       fmiTypeDefSize = 0;
    end
    //Size of VendorAnnotations
    positionOfVendor = find(xmlDescription.root.children.name == "VendorAnnotations")
    if positionOfVendor <> [] then
       fmiVendorAnSize = xmlDescription.root.children(positionOfVendor).children.size
    else
       fmiVendorAnSize = 0;
    end
    //Size of ModelVariables
    positionOfModelVar = find(xmlDescription.root.children.name == "ModelVariables")
    if positionOfModelVar <> [] then
       fmiModelVarSize = xmlDescription.root.children(positionOfModelVar).children.size
    else
       fmiModelVarSize = 0;
    end
    //Number of Direct Dependency
    fmiModelDependency = xmlXPath(xmlDescription, "/fmiModelDescription/ModelVariables/ScalarVariable/DirectDependency");
    //Default Experiment
    positionOfDefExp = find(xmlDescription.root.children.name == "DefaultExperiment")
    if positionOfDefExp <> [] then
       startTime = strtod(xmlDescription.root.children(positionOfDefExp).attributes.startTime);
       stopTime = strtod(xmlDescription.root.children(positionOfDefExp).attributes.stopTime);
       tolerance = strtod(xmlDescription.root.children(positionOfDefExp).attributes.tolerance);
    else
       startTime = []; stopTime = []; tolerance = [];
    end
    // fmi Type CS or ME
    fmiType = "Model Exchange";
    implList = [];
    positionOfCosimulation = find(xmlDescription.root.children.name == "CoSimulation");
      if positionOfCosimulation <> [] then
          fmiType = "CoSimulation"; // Version 2.0
          coSimulationRoot = xmlDescription.root.children(positionOfCosimulation);
          modelIdentifier = coSimulationRoot.attributes("modelIdentifier");
          implList = mlist([modelIdentifier, "Capabilities"], ...
          mlist(["Capabilities", ...
                 "canHandleVarCommunicStepSize", ...
                 "canInterpolateInputs", ...
                 "maxOutputDerivativeOrder", ...
                 "canRunAsynchronuously", ...
                 "canBeInstantiatedOnlyOncePerProcess", ...
                 "canNotUseMemoryManagementFunctions"], ...
                 coSimulationRoot.attributes("canHandleVariableCommunicationStepSize"), ...
                 coSimulationRoot.attributes("canInterpolateInputs"), ...
                 coSimulationRoot.attributes("maxOutputDerivativeOrder"), ...
                 coSimulationRoot.attributes("canRunAsynchronuously"), ...
                 coSimulationRoot.attributes("canBeInstantiatedOnlyOncePerProcess"), ...
                 coSimulationRoot.attributes("canNotUseMemoryManagementFunctions")));
      else
         modelIdentifier = xmlDescription.root.attributes.modelIdentifier;
         positionOfImpl = find(xmlDescription.root.children.name == "Implementation");
         if positionOfImpl <> [] then
           fmiType = "CoSimulation"; // Version 1.0

            // only one children is possible (CoSimulation_StandAlone or CoSimulation_Tool)
            if xmlDescription.root.children(positionOfImpl).children.size == 1 then
                implRoot = xmlDescription.root.children(positionOfImpl);
                if implRoot.children.size > 0 then
                    implName = implRoot.children(1).name;
                    implRootNew = implRoot.children(1).children(1);
                    implList = mlist([implName, "Capabilities"], ...
                               mlist(["Capabilities", ...
                                      "canHandleVarCommunicStepSize", ...
                                      "canHandleEvents", ...
                                      "canRejectSteps", ...
                                      "canInterpolateInputs", ...
                                      "maxOutputDerivativeOrder", ...
                                      "canRunAsynchronuously", ...
                                      "canSignalEvents", ...
                                      "canBeInstantiatedOnlyOncePerProcess", ...
                                      "canNotUseMemoryManagementFunctions"], ...
                                      implRootNew.attributes("canHandleVariableCommunicationStepSize"), ...
                                      implRootNew.attributes("canHandleEvents"), ...
                                      implRootNew.attributes("canRejectSteps"), ...
                                      implRootNew.attributes("canInterpolateInputs"), ...
                                      implRootNew.attributes("maxOutputDerivativeOrder"), ...
                                      implRootNew.attributes("canRunAsynchronuously"), ...
                                      implRootNew.attributes("canSignalEvents"), ...
                                      implRootNew.attributes("canBeInstantiatedOnlyOncePerProcess"), ...
                                      implRootNew.attributes("canNotUseMemoryManagementFunctions")));
              end
          if implName == "CoSimulation_Tool" & implRoot.children.size > 1 then
                   implRootNew = implRoot.children(1).children(2);
                   implRootNewSz = implRoot.children(1).children(2).children.size;
                   files = emptystr(implRootNewSz);
                   for i = 1:implRootNewSz
                       files(i) = implRoot.children(i).attributes;
                   end
                   setfield(3, mlist(["Model", ...
                                      "entryPoint", ...
                                      "manualStart", ...
                                      "type", ...
                                      "File"], ...
                                      implRoot.attributes.entryPoint, ...
                                      implRoot.attributes.entryPoint, ...
                                      implRoot.attributes.type, ...
                                      files), implList);
                elseif implName <> "CoSimulation_StandAlone" then
                    error(msprintf(_("%s: Wrong XML description.\n"),"fmiModelDescription"));
                end
            else
                error(msprintf(_("%s: Wrong XML description.\n"),"fmiModelDescription"));
            end
        end
    end

    //The attribute  “numberOfContinuousStates” available in FMI 1.0 has been removed for FMI 2.0,
    //since this information can be deduced from the remaining data in the XML file
    if xmlDescription.root.attributes.fmiVersion == "1.0" then
      numberOfContinuousStates  = strtod(xmlDescription.root.attributes.numberOfContinuousStates);
    else
      numberOfContinuousStates  = 0; //How to deduce it ??
    end
    description = mlist(["fmiDescript",...
                         "fmiVersion",...
                         "fmiType",...
                         "modelName",...
                         "modelIdentifier",...
                         "guid",...
                         "description"...
                         "author",...
                         "version",...
                         "generationTool",...
                         "generationDateAndTime",...
                         "variableNamingConvention",...
                         "numberOfContinuousStates",...
                         "numberOfEventIndicators",...
                         "UnitDefinitions",...
                         "TypeDefinitions",...
                         "DefaultExperiment",...
                         "VendorAnnotations",...
                         "ModelVariables",...
                         "DirectDependency",...
                         "Implementation"],...
                         xmlDescription.root.attributes.fmiVersion,...
                         fmiType,...
                         xmlDescription.root.attributes.modelName,...
                         modelIdentifier,...
                         xmlDescription.root.attributes.guid,...
                         xmlDescription.root.attributes.description,...
                         xmlDescription.root.attributes.author,...
                         xmlDescription.root.attributes.version,...
                         xmlDescription.root.attributes.generationTool,...
                         xmlDescription.root.attributes.generationDateAndTime,...
                         xmlDescription.root.attributes.variableNamingConvention,...
                         numberOfContinuousStates,...
                         strtod(xmlDescription.root.attributes.numberOfEventIndicators),...
                         fmiUnitDefSize,...
                         fmiTypeDefSize,...
                         mlist(["fmiExper", "startTime", "stopTime", "tolerance"],...
                         startTime, stopTime, tolerance),...
                         fmiVendorAnSize,...
                         fmiModelVarSize,...
                         fmiModelDependency.size,...
                         implList);

xmlDelete(xmlDescription);
endfunction
